'' =================================================================================================
''
''   File....... jm_textbuf.spin
''   Purpose.... Text buffer object for manipulating strings
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2015-2020 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon.mcphalen@gmail.com
''   Started....
''   Updated.... 25 AUG 2020
''
'' =================================================================================================


obj

  nstr : "jm_nstr"                                              ' numeric formatting routines
  

var

  long  p_buffer                                                ' address of buffer
  long  bsmax                                                   ' size of buffer
  byte  bidx                                                    ' buffer index (next char)
                                                                 
                                                                 
pub null                                                         
                                                                 
  ' This is not a top-level object                               


pub use(p_buf, maxlen)

'' Assign external buffer to object
'' -- p_buf is hub address of buffer
'' -- maxlen is maximum buffer length (without nul)
''    * external buffer must be declared as maxlen+1
'' -- does not reset buffer; allows on-the-fly switching

  longmove(@p_buffer, @p_buf, 2)                                ' copy pointer and max length
  bidx := strsize(p_buffer)                                     ' update bidx to end

                                                             
pub clear                                                        
                                                                 
'' Clears buffer and resets buffer index                         

  bytefill(p_buffer, 0, bsmax)                                  ' fill buffer with zeros
  bidx := 0


pub set(p_str) | n

'' Sets assigned buffer
'' -- replaces anything already in buffer
'' -- truncates input if required

  n := strsize(p_str) <# bsmax                                  ' number of bytes in source string

  bytemove(p_buffer, p_str, n)                                  ' move source string into buffer
  byte[p_buffer][n] := 0                                        ' terminate
  bidx := n                                                     ' update bidx


pub copy_to(p_str)

'' Copies the contents of buffer to array at p_str
'' -- output string must be sized hold contents of buffer

  bytemove(p_str, p_buffer, bidx+1)                             ' copy contents + 0
  

pub putc(c)

'' Add character to end of buffer

  if (bidx < bsmax)                                             ' if space in buffer
    byte[p_buffer][bidx++] := c                                 '  add new character
    byte[p_buffer][bidx] := 0                                   '  re-terminate
                                                                 
                                                                 
pub puts(p_str) | n                                                 
                                                                 
'' Add string to end of buffer
'' -- input string is terminated if too long for buffer

  n := strsize(p_str) <# (bsmax - bidx)                         ' get max chars to move

  bytemove(p_buffer+bidx, p_str, n)                             ' add string to buffer
  bidx += n                                                     ' update bidx
  byte[p_buffer][bidx] := 0                                     ' fix termination                                  


pub getc(idx)

'' Get character from idx'th position in buffer 

  if (idx < bidx)
    return byte[p_buffer][idx]
  else
    return 0
    

pub dec(value)

'' Add left-justified decimal to buffer

  puts(nstr.dec(value, 0))


pub fdec(value, digits)

'' Add fixed-width decimal to buffer
'' -- digits specifies maximum width

  puts(nstr.dec(value, digits))


pub jdec(value, digits, width, pad)

'' Add justified decimal to buffer
'' -- aligned in padded field (negative width to left-align)
'' -- digits is max number of digits to use
'' -- width is width of final field (max)
'' -- pad is character that fills out field

  puts(nstr.fmt_number(value, 10, digits, width, pad))


pub dpdec(value, dp)

'' Add decimal with decimal point to buffer
'' -- dp is number of digits after decimal point

  puts(nstr.dpdec(value, dp))
  

pub jdpdec(value, dp, width, pad)

'' Add justifed decimal with decimal point to buffer  
'' -- aligned in padded field (negative width to left-align)
'' -- dp is number of digits after decimal point
'' -- width is width of final field (max)
'' -- pad is character that fills out field

  puts(nstr.fmt_number(value, 99, dp, width, pad))


pub hex(value)

'' Add value as hexadecimal to buffer

  puts(nstr.po2toa(value, 16, 0))


pub fhex(value, digits)

'' Add value as hexadecimal using fixed # of digits to buffer

  puts(nstr.po2toa(value, 16, digits))


pub jhex(value, digits, width, pad)

'' Add value as quarternary using fixed # of digits to buffer
'' -- aligned inside field
'' -- pad fills out field

  puts(nstr.fmt_number(value, 16, digits, width, pad)) 


pub oct(value)

'' Add value as octal to buffer

  puts(nstr.po2toa(value, 8, 0))    


pub foct(value, digits)

'' Add value as octal using fixed # of digits to buffer

  puts(nstr.po2toa(value, 8, digits))

  
pub joct(value, digits, width, pad)

'' Add value as octal using fixed # of digits to buffer
'' -- aligned inside field
'' -- pad fills out field

  puts(nstr.fmt_number(value, 8, digits, width, pad)) 


pub qrt(value)

'' Add value as quarternary to buffer

  puts(nstr.po2toa(value, 4, 0))    


pub fqrt(value, digits)

'' Add value as quarternary using fixed # of digits to buffer

  puts(nstr.po2toa(value, 4, digits))


pub jqrt(value, digits, width, pad)

'' Add value as quarternary using fixed # of digits to buffer
'' -- aligned inside field
'' -- pad fills out field

  puts(nstr.fmt_number(value, 4, digits, width, pad))  


pub bin(value)

'' Add value as binary to buffer

  puts(nstr.po2toa(value, 2, 0))    


pub fbin(value, digits)

'' Add value as binary using fixed # of digits to buffer

  puts(nstr.po2toa(value, 2, digits))


pub jbin(value, digits, width, pad)

'' Add value as binary using fixed # of digits buffer
'' -- aligned inside field
'' -- pad fills out field

  puts(nstr.fmt_number(value, 2, digits, width, pad))


pub equal(p_str)

'' Return true if buffer is equal to string at p_str
'' -- case sensitive

  return strcomp(p_buffer, p_str)


pub nc_equal(p_str)

  return nc_strcomp(p_str)


pub nc_strcomp(p_str) | p_buf

'' Return true if buffer is equal to string at p_str
'' -- not case sensitive

  if (size <> strsize(p_str))                                   ' must be seme length
    return false

  p_buf := p_buffer

  repeat bidx
    if (upper(byte[p_str++]) <> upper(byte[p_buf++]))
      return false 
                                                      
  return true


pub str_index(p_list, lcount) | idx

'' Returns index of string (at p_str) in list of strings (at p_list)
'' -- lcount is number of strings in list
'' -- search is case sensitive

  idx := 0

  repeat lcount
    if (strcomp(p_buffer, p_list))
      return idx
    else
      ++idx
      p_list += strsize(p_list) + 1

  return -1
  

pub upper(c)

'' Convert c to upper case
'' -- does not affect buffer

  if ((c => "a") and (c =< "z"))
    c -= 32

  return c
  

pub ucase | idx

'' Converts alpha characters in buffer to upper case
                    
  repeat idx from 0 to bidx
    byte[p_buffer][idx] := upper(byte[p_buffer][idx])


pub lower(c)

'' Convert c to lower case
'' -- does not affect buffer

  if ((c => "A") and (c =< "Z"))
    c += 32

  return c

    
pub lcase | idx

'' Converts alpha characters in buffer to lower case
                    
  repeat idx from 0 to bidx
    byte[p_buffer][idx] := lower(byte[p_buffer][idx]) 


pub truncate(pos) | last

'' Sets index of buffer to pos
'' -- only if buffer longer than pos

  if ((pos => 0) and (pos < bsmax-1))
    bytefill(p_buffer+pos, 0, (bsmax+1)-pos)
    size
    return true

  return false


pub rtrim | idx                                  
                                                                 
'' Trims trailing whitespaces from string at p_str               

  repeat
    idx := size-1
    if ((idx =< 0) or (byte[p_buffer][idx] > " "))
      quit
    else
      byte[p_buffer][idx] := 0

 
pub backspace

'' Deletes last character in buffer
'' -- returns length of buffer

  if (bidx > 0)
    byte[p_buffer][--bidx] := 0                                 ' fixed 26 JUL 17

  return bidx

  
pub address

'' Exposes hub address of assigned buffer

  return p_buffer


pub str                                                         ' added for .str() methods

'' Provides buffer address for external string methods

  return p_buffer


pub size

'' Returns size of buffer

  bidx := strsize(p_buffer)                                     ' update

  return bidx


con

  ' Simple crypto
  ' -- for fun, not security!


pub mr_robot(p_str)                                            ' eps2.9_pyth0n-pt1.p7z

'' Encrypt/decrypt buffer using ROT-13 cypher
'' -- see http://www.decode.org/

  caesar(p_str, 13)
  

pub caesar(p_str, key) | idx, c

'' Encrypt/decrypt buffer using Caeser cipher
'' -- p_str is pointer to string to encode
''    * use -1 for buffer assigned to object
'' -- key is 1 to 26
''    * decryption key is 26-key

  if (p_str < 0)
    p_str := p_buffer                                           ' use assigned buffer

  idx := 0
    
  repeat strsize(p_str)
    c := byte[p_str][idx]
    if ((c => "A") and (c =< "Z"))
      c := "A" + ((c-"A") + key) // 26  
    elseif ((c => "a") and (c =< "z"))
      c := "a" + ((c-"a") + key) // 26  
    byte[p_str][idx++] := c    


pub rot47(p_str) | idx, c

'' Encrypt/decrypt buffer using ROT-47 cipher
'' -- p_str is pointer to string to encode
''    * use -1 for buffer assigned to object
'' -- includes punctuation and other characters

  if (p_str < 0)
    p_str := p_buffer                                           ' use assigned buffer

  idx := 0

  repeat strsize(p_str)
    c := byte[p_str][idx]  
    if ((c => 33) and (c =< 126))                               ' keyboard char
      c := 33 + ((c + 14) // 94)
      byte[p_str][idx] := c
    idx += 1    


pub exor(p_str, key) | idx, c

'' Encrypt/decrypt buffer by xor-ing str with key
'' -- p_str is pointer to string to encode
''    * use -1 for buffer assigned to object

  if (p_str < 0)
    p_str := p_buffer                                           ' use assigned buffer

  idx := 0
    
  repeat strsize(p_str)
    c := byte[p_str][idx]
    if ((c => "A") and (c =< "Z"))
      c ^= key  
    elseif ((c => "a") and (c =< "z"))
      c ^= key 
    byte[p_str][idx++] := c 
   
  
con { license }                                                  

{{

  Terms of Use: MIT License    

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:
  
  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.
  
  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

}}     